#include <stdio.h>
#define __USE_LARGEFILE64
#define __USE_FILE_OFFSET64
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include "rt_parser.h"

#define VERSION "1.0a2 4-03-2004"

/* header for a Queue */

struct queue_hdr {
	struct   buf_hdr     *q_head_ptr   ; /* proc Q head             */
	struct   buf_hdr     *q_tail_ptr   ; /* proc Q tail             */
	int                  q_tot_bytes   ; /* total table size        */
	short                q_valid       ; /* 1 - table valid         */
					     /* 0 - table NOT valid     */
	short                q_ver         ; /* 1 - table valid         */
};

#define MAX_NAME_LENGTH		512
#define DEFAULT_INPUT_FILE		"/dev/dtv"
#define OUTPUT_NAME_EXTENSION	".m2t"
#define TRANSPORT_PKT_SIZE		188
// buffer need not be a multiple of the packet size (but is more efficient)
// buffer must be large enough for at least 3 packets (for sync)
#define INPUT_BUFFER_SIZE		TRANSPORT_PKT_SIZE * 64
// output buffer should be a multiple of the packet size
#define OUTPUT_BUFFER_SIZE		TRANSPORT_PKT_SIZE * 64
#define SYNC_BYTE				0x47
#define SYNC_SEARCH_BYTE_COUNT	TRANSPORT_PKT_SIZE * 4000
#define DEFAULT_SCAN_COUNT		20000
#define MAX_SELECTED_STREAMS	20

extern int errno;

char input_file[MAX_NAME_LENGTH] = "";
char output_file[MAX_NAME_LENGTH] = "";
int in_fd, out_fd;
char in_open = 0;
char out_open = 0;
mode_t open_mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
long long buffer_file_offset;
long long lost_sync_offset;
int do_read_init;
char write_replace = 0;
char debug_output = 0;
char verbose_output = 0;
char info_only = 0;
#ifdef CHANNEL_TUNING
int channel_number = 0;
#endif
int program_number = 0;
int minutes_limit = 0;
int seconds_limit = 0;
int total_limit;
char use_default_program = 0;
char use_all_packets = 0;
int scan_count = DEFAULT_SCAN_COUNT;
unsigned int selected_streams[MAX_SELECTED_STREAMS];		
int selected_streams_count;
char too_many_streams;
int input_is_char_dev;

unsigned char *input_buffer;
unsigned char *input_ptr;
unsigned char *input_end;

unsigned char *output_buffer;
unsigned char *output_ptr;

char input_eof;
int default_program = 0;

unsigned char pat_packet[TRANSPORT_PKT_SIZE];
unsigned char transport_stream_id[2];
int packet_counter = 0;

int got_interrupt;
int got_alarm;
unsigned long long processed_packet_count;
unsigned long long write_packet_count;

static void sig_handler(int signo);
int scan_packets(void);
unsigned int calc_crc(unsigned char *buf, unsigned int len);
int process_packets(void);
int process_one_packet(unsigned char *packet);
int output_pat_packet (unsigned char *packet);
int output_packet(unsigned char *packet);
int flush_output(void);
int get_new_packet(unsigned char **packet);
int fill_input_buffer(void);
int buffer_sync_ok(void);
void print_help(char *myname);
int convert_to_int(char *str, int *convert_error);
char * safe_strcat(char *dest, char *src, int src_copylen, int maxlen);
void close_files(void);
void add_stream(int stream);
int prompt_program(void);
void add_pmt_streams(struct queue_hdr *q_hdr);
void print_pmt(int prog_number, struct queue_hdr *q_hdr);
void get_channel_desc(int prog_number, char *channel_desc);
void get_lang(int prog_number, int pid_id, char *lang);

#ifdef CHANNEL_TUNING
extern int tune_to_channel(int tuner_fd, int channel_number);
#endif

extern void process_packet(char *packet_ptr, char do_init);
extern void prt_summary(void);
extern struct pat_hdr pat_hdr;
extern struct vct_hdr current_tvct_hdr;
extern unsigned long int g_vec;
struct queue_hdr *eit_queue = 0;

int main(int argc, char *argv[])
{
	char *arg;
	char *myname;
	char in_opts = 1;
	char argchar;
	int convert_error;
	char *afterslash, *enddot;
	int len;
	int saved_errno;
	int result;
	mode_t open_flags;
	off_t seek_result;
	struct stat in_stat;
	
	arg = *argv++;
	argc--;
	myname = arg;
	
	while (argc > 0) {
		arg = *argv++;
		argc--;

		// process options
		if (strcmp(arg, "--help") == 0) {
			print_help(myname);
			return 0;
			}
		if ((arg[0] == '-') && in_opts) {
			// need exactly one letter
			if (arg[1] == 0) {
				fprintf(stderr, "Missing option after \"-\".\n");
				print_help(myname);
				return 1;
				}
			else if (arg[2] != 0) {
				fprintf(stderr, "Invalid option: \"%s\".\n", arg);
				print_help(myname);
				return 1;
				}
			// process options
			argchar = arg[1];
			switch(argchar) {
				// "--" ends option processing
				case '-':
						in_opts = 0;
						break;
						
				// options with arguments
				case 'a':
#ifdef CHANNEL_TUNING
				case 'c':
#endif
				case 'i':
				case 'm':
				case 'o':
				case 'p':
				case 's':
						if (argc > 0) {
							arg = *argv++;
							argc--;
							switch(argchar) {
								case 'a':
									scan_count = convert_to_int(arg, &convert_error);
									if ((convert_error != 0) || (scan_count < 0)) {
										fprintf(stderr, "Invalid argument \"%s\" for -a.\n", arg);
										print_help(myname);
										return 1;
										}
									break;
#ifdef CHANNEL_TUNING
								case 'c':
									channel_number = convert_to_int(arg, &convert_error);
									if ((convert_error != 0) || (channel_number < 2) || (channel_number > 83)) {
										fprintf(stderr, "Invalid argument \"%s\" for -c.\n", arg);
										print_help(myname);
										return 1;
										}
									break;
#endif
								case 'i':
									input_file[0] = 0;
									safe_strcat(input_file, arg, -1, MAX_NAME_LENGTH);
									break;
								case 'm':
									minutes_limit = convert_to_int(arg, &convert_error);
									if ((convert_error != 0) || (minutes_limit < 0)) {
										fprintf(stderr, "Invalid argument \"%s\" for -m.\n", arg);
										print_help(myname);
										return 1;
										}
									break;
								case 'o':
									output_file[0] = 0;
									safe_strcat(output_file, arg, -1, MAX_NAME_LENGTH);
									break;
								case 'p':
									if (strcmp(arg, "default") == 0) {
										use_default_program = 1;
										break;
										}
									if (strcmp(arg, "all") == 0) {
										use_all_packets = 1;
										break;
										}
									program_number = convert_to_int(arg, &convert_error);
									if ((convert_error != 0) || (program_number < 1)) {
										fprintf(stderr, "Invalid argument \"%s\" for -p.\n", arg);
										print_help(myname);
										return 1;
										}
									break;
								case 's':
									seconds_limit = convert_to_int(arg, &convert_error);
									if ((convert_error != 0) || (seconds_limit < 0)) {
										fprintf(stderr, "Invalid argument \"%s\" for -s.\n", arg);
										print_help(myname);
										return 1;
									break;
								default:
									break;
									}
								}
							}							
						else {
							fprintf(stderr, "Argument missing for -%c.\n", arg[1]);
							print_help(myname);
							return 1;
							}
						break;
						
				case 'q':
						info_only = 1;
						break;

				case 'r':
						write_replace = 0;
						break;

				case 'R':
						write_replace = 1;
						break;

				case 'd':
						debug_output = 1;
						break;
						
				case 'v':
						verbose_output = 1;
						break;
						
				case 'h':
						print_help(myname);
						return 0;
						break;
						
				default:
						fprintf(stderr, "Invalid option: \"%s\".\n", arg);
						print_help(myname);
						return 1;
						break;
				}

			}

		// error if an argument specified
		else {
			fprintf(stderr, "Invalid argument: \"%s\" -- only options are used.\n", arg);
			print_help(myname);
			return 1;
			}
		}		
		
	// debug implies verbose
	if (debug_output) verbose_output = 1;

	// set file names not specified
	// set default input
	if (input_file[0] == 0) {
		strcpy(input_file, DEFAULT_INPUT_FILE);
		}

	// default output file is input_name.m2t
	if (output_file[0] == 0) {
		// copy input filename up to last period (if any)
		// for /dev get input file name following last /
		if (strncmp(input_file, "/dev/", 5) == 0) {
			afterslash = strrchr(input_file, '/');
			if (afterslash == 0) {
				afterslash = input_file;
				}
			else {
				afterslash++;
				}
			}
		else {
			afterslash = input_file;
			}
		
		len = 0;
		// don't use input file name after last period
		enddot = strrchr(afterslash, '.');
		if (enddot != 0) {
			len = enddot - afterslash;
			}
		if (len <= 0) {
			len = strlen(afterslash);
			}
		safe_strcat(output_file, afterslash, len,
					MAX_NAME_LENGTH - strlen(OUTPUT_NAME_EXTENSION));
		safe_strcat(output_file, OUTPUT_NAME_EXTENSION, -1, MAX_NAME_LENGTH);
		}
	
	// compute total time limit
	total_limit = seconds_limit + minutes_limit * 60;

	// for debugging, display files and options
	if (debug_output) {
		printf("Settings:\n");
		printf("\tInput file = \"%s\", Output file = \"%s\", Replace output = %d\n",
			   input_file, output_file, write_replace);
		printf("\tMinutes limit = %d, Seconds limit = %d, Total seconds limit = %d\n",
			   minutes_limit, seconds_limit, total_limit);	
#ifdef CHANNEL_TUNING
		printf("\tChannel = %d, Program = %d, Packet scan count = %d\n",
			   channel_number, program_number, scan_count);
#else
		printf("\tProgram = %d, Packet scan count = %d\n",
			   program_number, scan_count);
#endif
		printf("\tDebug output = %d, Verbose output = %d\n",
			   debug_output, verbose_output);
		}

	// open input file
	open_flags = O_RDONLY;
#ifdef O_LARGEFILE
	open_flags |= O_LARGEFILE;
	if (debug_output) printf("Using O_LARGEFILE for open.\n");
#else
	if (debug_output) printf("Not using O_LARGEFILE for open.\n");
#endif
	in_fd = open(input_file, open_flags);
	if (in_fd == -1) {
		fprintf(stderr, "Unable to open \"%s\": %s\n", input_file, strerror(errno));
		return(errno);
		}
	in_open = 1;

	// determine if the input file is a character device
	// (i.e. our tuner)
	result = fstat(in_fd, &in_stat);
	if (result != 0) {
		fprintf(stderr, "Unable to stat \"%s\": %s\n", input_file, strerror(errno));
		close_files();
		return(errno);
		}
	input_is_char_dev = S_ISCHR(in_stat.st_mode);

#ifdef CHANNEL_TUNING
	// if a channel number was specified, attempt to tune to it
	if (channel_number != 0) {
		if (!input_is_char_dev) {
			fprintf(stderr, "Input file does not permit changing channels- not a character device.\n");
			close_files();
			return 2;
			}
		result = tune_to_channel(in_fd, channel_number);
		if (result != 0) {
			fprintf(stderr, "Unable to tune \"%s\" to channel %d: %s\n",
					input_file, channel_number, strerror(errno));
			close_files();
			return 2;
			}
		else if (verbose_output) {
			printf("Tuned \"%s\" to channel %d.\n", input_file, channel_number);
			}
		}
#endif

	// allocate the input buffer
	input_buffer = (unsigned char *)malloc(INPUT_BUFFER_SIZE);
	if (input_buffer == 0) {
		fprintf(stderr, "Unable to allocate input buffer of size %u\n", INPUT_BUFFER_SIZE);
		close_files();
		return ENOMEM;
		}

	// init input buffer
	do_read_init = 1;

	// if we're not processing all packets, scan for program information
	if (!use_all_packets) {
		result = scan_packets();
		if (result != 0) {
			goto cleanup;
			}
		}

	// done if only info requested
	if (info_only) {
		result = 0;
		goto cleanup;
		}

	if (verbose_output && use_all_packets) {		// all may be set by scan
		printf("Processing all packets.\n");
		}

	// need to re-open input file if did scan
	if (!in_open) {
		open_flags = O_RDONLY;
#ifdef O_LARGEFILE
		open_flags |= O_LARGEFILE;
#endif
		in_fd = open(input_file, open_flags);
		if (in_fd == -1) {
			fprintf(stderr, "Unable to open \"%s\": %s\n", input_file, strerror(errno));
			result = errno;
			goto cleanup;
			}
		in_open = 1;

		// reset input buffer
		do_read_init = 1;
		}

	// open output file
	open_flags = O_WRONLY | O_CREAT;
	if (write_replace == 0) open_flags |= O_EXCL;
	else open_flags |= O_TRUNC;
#ifdef O_LARGEFILE
	open_flags |= O_LARGEFILE;
#endif
	out_fd = open(output_file, open_flags, open_mode);
	if (out_fd == -1) {
		fprintf(stderr, "Unable to create \"%s\": %s\n", output_file, strerror(errno));
		result = errno;
		goto cleanup;
		}
	out_open = 1;

	// allocate output buffer
	output_buffer = (unsigned char *)malloc(OUTPUT_BUFFER_SIZE);
	if (output_buffer == 0) {
		fprintf(stderr, "Unable to allocate output buffer of size %u\n", OUTPUT_BUFFER_SIZE);
		result = ENOMEM;
		goto cleanup;
		}
	output_ptr = output_buffer;

	// process
	result = process_packets();
	free(output_buffer);
	
cleanup:
	// clean up and return
	free(input_buffer);
	close_files();

	return result;
}

static void sig_handler(int signo)
{
	switch (signo) {
		case SIGINT:
			got_interrupt = 1;
			printf("Terminating due to user interrupt.\n");
			break;
			
		case SIGALRM:
			got_alarm = 1;
			printf("Terminating at end of time limit.\n");
			break;
		
		default:
			break;		
		}
}

int scan_packets(void)
{
unsigned char *packet;
int result;
static char init = 1;
static int count = 0;
struct pat_item *item_ptr;
int i, streams;
char channel_desc[32];
int program_to_use = 0;
int prompt_result = 0;
unsigned int ts_pid;
unsigned int pmt_pid;
unsigned int crc;

transport_stream_id[0] = 0;
transport_stream_id[1] = 0;

result = get_new_packet(&packet);
while ((result == 0) && (count < scan_count)) {
	ts_pid = ((packet[1] & 0x1f) << 8) | (packet[2] & 0xff);
	if (ts_pid == 0) {	// save transport stream id
		transport_stream_id[0] = packet[8];
		transport_stream_id[1] = packet[9];
		}
	process_packet(packet, init);
	init = 0;
	result = get_new_packet(&packet);
	count++;
	}

// close input file
if (in_open) {
	close(in_fd);
	in_open = 0;
	}

if (count == 0) {
	fprintf(stderr, "No valid packets found.\n");
	return 1;
	}
	
if (count < scan_count) {
	fprintf(stderr, "Only able to scan %d packets instead of %d\n",
			count, scan_count);
	}

if (verbose_output) {
	printf("Summary after scanning first %d packets:\n", scan_count);
	prt_summary();
	}
	

// Count number of valid programs
streams = 0;
if (pat_hdr.num_items > 0) {
	item_ptr = pat_hdr.pat_item_ptr;
	// check PMT table exists for each stream
	for (i = 0; i < pat_hdr.num_items; i++) {
		if ((item_ptr->pmt_q_hdr_ptr)->q_head_ptr != 0) {
			streams++;
			if (default_program == 0) default_program = item_ptr->program_number;
			}
		item_ptr++;
		}
	}
	
if (verbose_output) {
	printf("Stream contains %d total program(s).\n", pat_hdr.num_items);
	}
if (streams > 0) {
	printf("Stream contains %d valid program(s).\n", streams);
	}
else {
	fprintf(stderr, "No valid programs found in stream.\n");
	return 2;
	}

if (streams == 1) {
	if (program_number == 0) {
		program_number = default_program;
		}
	else if (program_number != default_program) {
		fprintf(stderr, "Requested program %d not found in stream.\n", program_number);
		return 3;
		}
	}

// Print each program and streams
item_ptr = pat_hdr.pat_item_ptr;
for (i = 0; i < pat_hdr.num_items; i++, item_ptr++) {
	if ((item_ptr->pmt_q_hdr_ptr)->q_head_ptr != 0) {
		if (program_number != 0) {
			if (item_ptr->program_number != program_number) {
				continue;
				}
			else {
				printf("    Using program %d:", item_ptr->program_number);
				program_to_use = program_number;
				}
			}
		else {
			printf("    Program %d:", item_ptr->program_number);
			}
		// note: get_channel_desc will set default_program to minor program 1
		get_channel_desc(item_ptr->program_number, channel_desc);
		if (channel_desc[0] == 0) {
			printf("\n");
			}
		else {
			printf(" %s:\n", channel_desc);
			}
		print_pmt(item_ptr->program_number, item_ptr->pmt_q_hdr_ptr);
		}
	}

if (info_only) return 0;

if (program_to_use == 0) {
	if (program_number != 0) {
		fprintf(stderr, "Requested program %d not found in stream.\n", program_number);
		return 3;
		}
	if (use_default_program) {
		program_to_use = default_program;
		}
	else {
	new_prompt:
		prompt_result = prompt_program();
		if (use_all_packets) return 0;
		if (prompt_result == 0) return 4;
		// check returned number is a program number
		item_ptr = pat_hdr.pat_item_ptr;
		for (i = 0; i < pat_hdr.num_items; i++) {
			if ((item_ptr->pmt_q_hdr_ptr)->q_head_ptr != 0) {
				if (item_ptr->program_number == prompt_result) {
					program_to_use = prompt_result;
					}
				}
			item_ptr++;
			}
		if (program_to_use == 0) {
			printf("%d is not a valid program number\n", prompt_result);
			goto new_prompt;		
			}
		}
	}

if (verbose_output) {
	printf("Selecting streams for program %d\n", program_to_use);
	}

// define streams to output
selected_streams_count = 0;
too_many_streams = 0;

// PAT stream is not added- implicitly processed

// add the streams for the selected program
item_ptr = pat_hdr.pat_item_ptr;
for (i = 0; i < pat_hdr.num_items; i++) {
	if ((item_ptr->pmt_q_hdr_ptr)->q_head_ptr != 0) {
		if (item_ptr->program_number == program_to_use) {
			pmt_pid = item_ptr->program_map_pid;
			add_stream(pmt_pid);
			add_pmt_streams(item_ptr->pmt_q_hdr_ptr);
			break;
			}
		}
	item_ptr++;
	}

if (debug_output) {
	printf("%d streams selected:\n", selected_streams_count);
	for (i = 0; i < selected_streams_count; i++) {
		printf("    %02x\n", selected_streams[i]);
		}
	}

if (too_many_streams) {
	fprintf(stderr, "Stream table is too small for selected streams.\n");
	return 5;
	}

// generate the new PAT packet with just the selected streams

//initialize packet with 0xff
memset(pat_packet, 0xff, TRANSPORT_PKT_SIZE);

// set fields
pat_packet[0] = 0x47;		// sync
pat_packet[1] = 0x40;		// PUSI etc, PID
pat_packet[2] = 0;			// rest of PID
pat_packet[3] = 0x10;		// adaption etc, counter
pat_packet[4] = 0;			// offset
pat_packet[5] = 0;			// table id
pat_packet[6] = 0xb0;		// fixed bits, len
pat_packet[7] = 13;			// rest of len
pat_packet[8] = transport_stream_id[0];
pat_packet[9] = transport_stream_id[1];
pat_packet[10] = 0xc1;		// reserved, version, current
pat_packet[11] = 0;			// section number
pat_packet[12] = 0;			// last section number
pat_packet[13] = (program_to_use & 0x0000ff00) >> 8;
pat_packet[14] = program_to_use & 0x000000ff;
pat_packet[15] = 0xe0 | ((pmt_pid & 0x00001f00) >> 8);
pat_packet[16] = pmt_pid & 0x000000ff;
// calculate and store CRC
crc = calc_crc(pat_packet+5, 12);
pat_packet[17] = (crc & 0xff000000) >> 24;
pat_packet[18] = (crc & 0x00ff0000) >> 16;
pat_packet[19] = (crc & 0x0000ff00) >> 8;
pat_packet[20] = crc & 0x000000ff;

return 0;
}

unsigned int calc_crc(unsigned char *buf, unsigned int len)
{
   int               j;
   unsigned int      length;
   unsigned char     data_bit, next_byte;
   unsigned long int shift_reg;

   shift_reg = 0xFFFFFFFF;
   length = len;

    while ( length-- > 0 )
    {
	 next_byte = *buf++;
	 for (j=7; j>=0; j--)
	 {
	    data_bit = (next_byte>>j)&0x01;

	    data_bit ^= ((shift_reg>>31)&0x01);
	    shift_reg = (shift_reg<<1);
	    if ( data_bit == 1 )
	    {
		  shift_reg = (shift_reg ^ g_vec) | 1 ;
	    }
	 }
    }
    return( shift_reg );
}

void add_stream(int stream)
{
int i;

// room for stream?
if (selected_streams_count == MAX_SELECTED_STREAMS) {
	too_many_streams = 1;
	return;
	}

// if stream is already present, don't add again
if (selected_streams_count > 0) {
	for (i = 0; i < selected_streams_count; i++) {
		if (selected_streams[i] == stream) {
			return;
			}
		}
	}

selected_streams[selected_streams_count++] = stream;
}

int prompt_program(void)
{
char input_line[256];
int convert_error, result;

retry:
	printf("Enter program number to process, \"all\", or \"quit\":\n");
	scanf("%s", input_line);

	if (strcmp(input_line, "all") == 0) {
		use_all_packets = 1;
		return 0;
		}

	if (strcmp(input_line, "quit") == 0) {
		return 0;
		}

	result = convert_to_int(input_line, &convert_error);
	if (convert_error) {
		printf("\"%s\" is not a valid number\n", input_line);
		goto retry;
		}
	return result;
}

void add_pmt_streams(struct queue_hdr *q_hdr)
{
    char *buf_ptr, *c_ptr, *descr_ptr;
    unsigned int tbl_sect_len;
    unsigned int len_left, descr_len;
    struct buf_hdr *buf_hdr_ptr;
    unsigned int pid_id;
    unsigned int pcr_pid_id;	// Program Clock Reference stream

    buf_hdr_ptr = q_hdr->q_head_ptr;
    buf_ptr = buf_hdr_ptr->buf_ptr;

	pcr_pid_id = ((buf_ptr[8]&0x1F)<<8) | (buf_ptr[9]&0xFF);
	if (pcr_pid_id != 0x1fff) add_stream(pcr_pid_id);
	tbl_sect_len = ((buf_ptr[1]&0x0F)<<8) | (buf_ptr[2]&0xFF);
	descr_len = ((buf_ptr[10]&0x0F)<<8) | (buf_ptr[11]&0xFF) ;
	descr_ptr =  buf_ptr + 12;
	c_ptr = buf_ptr + descr_len + 12;
	len_left = (tbl_sect_len - 9 - 4) - (descr_len) ;
	while ( len_left > 0 ) {
		pid_id = ((c_ptr[1]&0x1F)<<8) | (c_ptr[2]&0xFF);
		add_stream(pid_id);
	    descr_len = ((c_ptr[3]&0x0F)<<8) | (c_ptr[4]&0xFF);
	    descr_ptr = c_ptr+5;
	    c_ptr += (descr_len + 5) ;
	    len_left -= (descr_len + 5) ;
		}
}

void print_pmt(int prog_number, struct queue_hdr *q_hdr)
{
    char *buf_ptr, *c_ptr, *descr_ptr;
    unsigned int tbl_sect_len;
    unsigned int len_left, descr_len;
    struct buf_hdr *buf_hdr_ptr;
    char stream_name[128];
    char stream_lang[32];
    int pid_id;

    buf_hdr_ptr = q_hdr->q_head_ptr;
	if (buf_hdr_ptr == 0) {
		fprintf(stderr, "\tNo PMT table exists!\n");
		return;
		}
    buf_ptr = buf_hdr_ptr->buf_ptr;

    tbl_sect_len = ((buf_ptr[1]&0x0F)<<8) | (buf_ptr[2]&0xFF);
	descr_len = ((buf_ptr[10]&0x0F)<<8) | (buf_ptr[11]&0xFF) ;
	descr_ptr =  buf_ptr + 12;
	c_ptr = buf_ptr + descr_len + 12;
	len_left = (tbl_sect_len - 9 - 4) - (descr_len) ;
	while ( len_left > 0 )
	{
		pid_id = ((c_ptr[1]&0x1F)<<8) | (c_ptr[2]&0xFF);
		switch (c_ptr[0] & 0xff) {
			case 0x01:
					strcpy(stream_name, "Video (MPEG-1)");
					get_lang(prog_number, pid_id,  stream_lang);
					if (stream_lang[0] != 0) {
						sprintf(stream_name + strlen(stream_name),
							", language = %s", stream_lang);
						}
					break;
			case 0x02:
					strcpy(stream_name, "Video (MPEG-2)");
					get_lang(prog_number, pid_id,  stream_lang);
					if (stream_lang[0] != 0) {
						sprintf(stream_name + strlen(stream_name),
							", language = %s", stream_lang);
						}
					break;
			case 0x03:
					strcpy(stream_name, "Audio (MPEG-1)");
					get_lang(prog_number, pid_id,  stream_lang);
					if (stream_lang[0] != 0) {
						sprintf(stream_name + strlen(stream_name),
							", language = %s", stream_lang);
						}
					break;
			case 0x04:
					strcpy(stream_name, "Audio (MPEG-2)");
					get_lang(prog_number, pid_id,  stream_lang);
					if (stream_lang[0] != 0) {
						sprintf(stream_name + strlen(stream_name),
							", language = %s", stream_lang);
						}
					break;
			case 0x81:
					strcpy(stream_name, "Audio (Dolby AC-3)");
					get_lang(prog_number, pid_id,  stream_lang);
					if (stream_lang[0] != 0) {
						sprintf(stream_name + strlen(stream_name),
							", language = %s", stream_lang);
						}
					break;
			default:
					sprintf(stream_name, "Unknown (0x%02x)", c_ptr[0] & 0xff);
					break;
			}
		
	    printf("        %s\n", stream_name);
	    descr_len = ((c_ptr[3]&0x0F)<<8) | (c_ptr[4]&0xFF);
	    descr_ptr = c_ptr+5;
	    c_ptr += (descr_len + 5) ;
	    len_left -= (descr_len + 5) ;
	}
}


void get_channel_desc(int prog_number, char *channel_desc)
{
	int i, j;
	struct vct_hdr *current_vct_hdr;
	struct vct_item *vct_item_ptr;
	unsigned char short_name[8];
	
	channel_desc[0] = 0;
	
	current_vct_hdr = &current_tvct_hdr;
	vct_item_ptr = current_vct_hdr->vct_tbl_ptr;
	if (vct_item_ptr == 0) return;
	for (i = 0; i < current_vct_hdr->num_entries; i++, vct_item_ptr++) {
		if (vct_item_ptr->program_number == prog_number) {
			for (j = 0; j < 7; j++) {
				short_name[j] = vct_item_ptr->short_name[j];
				}
			short_name[7] = 0;			
			sprintf(channel_desc, "%d-%d %s",
				vct_item_ptr->major_channel_number,
				vct_item_ptr->minor_channel_number,
				short_name);
			if (vct_item_ptr->minor_channel_number == 1) {
				default_program = prog_number;
				}
			break;
			}		
		}
}

void get_lang(int prog_number, int pid_id, char *lang)
{
	int i, j, k;
	struct vct_hdr *current_vct_hdr;
	struct vct_item *vct_item_ptr;
	unsigned char lang_name[4];
	
	lang[0] = 0;
	current_vct_hdr = &current_tvct_hdr;
	vct_item_ptr = current_vct_hdr->vct_tbl_ptr;
	if (vct_item_ptr == 0) return;
	for (i = 0; i < current_vct_hdr->num_entries; i++, vct_item_ptr++) {
		if (vct_item_ptr->program_number == prog_number) {
			for (k=0; k < vct_item_ptr->num_ch_elts; k++) {
				if (vct_item_ptr->elts[k].pes_pid == pid_id) {
					for (j = 0; j < 3; j++) {
						lang_name[j] = vct_item_ptr->elts[k].code_639[j];
						}
					lang_name[3] = 0;
					strcpy(lang, lang_name);				
					break;
					}
				}
			break;
			}		
		}
}


int process_packets(void)
{
unsigned char *packet;
int read_result, write_result;
static char init = 1;
static int count = 0;
struct pat_item *item_ptr;
int i, streams;
char channel_desc[32];
int program_to_use = 0;
int prompt_result = 0;
time_t start_time, elapsed_time;
int rate, hours, mins, secs;

if (signal(SIGINT, sig_handler) == SIG_ERR) {
	fprintf(stderr, "Error setting SIGINT handler\n");
	return 1;
	}

if (signal(SIGALRM, sig_handler) == SIG_ERR) {
	fprintf(stderr, "Error setting SIGALRM handler\n");
	return 1;
	}

write_result = 0;
got_interrupt = 0;
got_alarm = 0;
if (total_limit > 0) {
	if (!input_is_char_dev) {
		fprintf(stderr, "Warning: Setting time limit when input is not a device.\n");
		}
	alarm(total_limit);
	}

processed_packet_count = 0;
write_packet_count = 0;
start_time = time(0);
read_result = get_new_packet(&packet);
while ((read_result == 0) && (write_result == 0) 
       && (got_interrupt == 0) && (got_alarm == 0)) {
	processed_packet_count++;
	write_result = process_one_packet(packet);
	if (write_result == 0) {
		read_result = get_new_packet(&packet);
		}
	}

if (write_result == 0) {
	write_result = flush_output();
	}

elapsed_time = time(0) - start_time;
if (elapsed_time > 0) {
	rate = processed_packet_count / elapsed_time;
	}
else {
	rate = 0;
	}

hours = elapsed_time / 3600;
elapsed_time -= hours * 3600;
mins = elapsed_time / 60;
elapsed_time -= mins * 60;
secs = elapsed_time;

printf("Processed %lld packets, Wrote %lld packets to \"%s\".\n",
	processed_packet_count, write_packet_count, output_file);
	
printf("Elapsed time = %02d:%02d:%02d, Rate = %d packets/second.\n",
	hours, mins, secs, rate);

if (input_eof) {
	read_result = 0;
	}

if (write_result != 0) {
	return write_result;
	}
else {
	return read_result;
	}
}

int process_one_packet(unsigned char *packet)
{
unsigned int ts_pid;
int i, result;

result = 0;

// if doing selected packets, check if this one should be included
if (use_all_packets) {
	result = output_packet(packet);
	}
else {
	ts_pid = ((packet[1] & 0x1f) << 8) | (packet[2] & 0xff);
	if (ts_pid == 0) result = output_pat_packet(packet);
	else {
		for (i = 0; i < selected_streams_count; i++) {
			if (selected_streams[i] == ts_pid) {
				result = output_packet(packet);
				break;
				}		
			}
		}
	}

return result;
}

int output_pat_packet (unsigned char *packet)
{
	int current, section;
	
	// skip packets for additional sections
	section = packet[11];
	if (section != 0) return 0;
	
	// skip packet which is not current
	current = packet[10] & 0x01;
	if (current == 0) return 0;

	// store counter value in packet
	pat_packet[3] = 0x10 | (packet_counter & 0x0f);
	packet_counter++;
	if (packet_counter == 16) packet_counter = 0;

	// output our PAT packet
	return(output_packet(pat_packet));
}

int output_packet(unsigned char *packet)
{
	ssize_t used, remain, write_actual;
	
	used = output_ptr - output_buffer;
	remain = OUTPUT_BUFFER_SIZE - used;
	
	// if packet will fit in buffer, copy it and return
	if (remain >= TRANSPORT_PKT_SIZE) {
		memcpy(output_ptr, packet, TRANSPORT_PKT_SIZE);
		output_ptr += TRANSPORT_PKT_SIZE;
		return 0;
		}	

	// need to write out the buffer first
	write_actual = write(out_fd, output_buffer, used);
	if (write_actual != used) {
		fprintf(stderr, "Error writing output file %s: %s\n", output_file, strerror(errno));
		return errno;
		}	
	write_packet_count += used / TRANSPORT_PKT_SIZE;

	// now can store the packet in the buffer
	memcpy(output_buffer, packet, TRANSPORT_PKT_SIZE);
	output_ptr = output_buffer + TRANSPORT_PKT_SIZE;
	return 0;
}

int flush_output(void)
{
	ssize_t used, write_actual;
	
	used = output_ptr - output_buffer;
	if (used > 0) {	
		write_actual = write(out_fd, output_buffer, used);
		if (write_actual != used) {
			fprintf(stderr, "Error writing output file %s: %s\n", output_file, strerror(errno));
			return errno;
			}	
		write_packet_count += used / TRANSPORT_PKT_SIZE;
		}	

	return 0;
}

int get_new_packet(unsigned char **packet)
{
int remain;
int result;
unsigned int sync_count;

if (do_read_init) {
	input_ptr = input_buffer;
	input_end = input_buffer;
	input_eof = 0;
	remain = 0;
	}
else {
	// get remaining length in buffer
	remain = input_end - input_ptr;
	remain -= TRANSPORT_PKT_SIZE;			// did this packet
	}	
	
// if not long enough, refill buffer
if (remain < TRANSPORT_PKT_SIZE) {
	// return done if last read reached eof
	if (input_eof) return -1;

	// move any remaining data to the start of the buffer
	if (remain > 0) {
		input_ptr += TRANSPORT_PKT_SIZE;
		memmove(input_buffer, input_ptr, remain);
		input_end = input_buffer + remain;
		}
	else {
		input_end = input_buffer;
		}
	if (do_read_init) {
		buffer_file_offset = 0;		// first read
		do_read_init = 0;
		}
	else {
		buffer_file_offset += INPUT_BUFFER_SIZE - remain;
		}
	result = fill_input_buffer();
	if (result != 0) {
		return result;
		}
	remain = input_end - input_buffer;
	if (remain < TRANSPORT_PKT_SIZE) {
		return -1;
		}
	input_ptr = input_buffer;
	}
else {
	input_ptr += TRANSPORT_PKT_SIZE;
	}	

// if packet is valid, ready to return
if (*input_ptr == SYNC_BYTE) {
	*packet = input_ptr;
	return 0;
	}

// need to resynchronize
lost_sync_offset = buffer_file_offset + (input_ptr - input_buffer);
if ((lost_sync_offset != 0) || (verbose_output)) {
	fprintf(stderr, "Lost sync at input offset %lld\n", lost_sync_offset);
	}

sync_count = 0;
while ((!buffer_sync_ok()) && (sync_count < SYNC_SEARCH_BYTE_COUNT)) {
	// at this point, input_ptr is at the start of one or more full
	// packets, but the starting point is out of sync

	// try to move input_ptr ahead to another sync byte
	remain = input_end - input_ptr - 1;
	input_ptr = memchr(input_ptr + 1, SYNC_BYTE, remain);
	sync_count += remain;
	if (input_ptr == 0) input_ptr = input_end;
	remain = input_end - input_ptr;		// 0 if sync not found
	// try to get at least 3 packets to check
	if (remain < (TRANSPORT_PKT_SIZE * 3)) {
		// discard data before possible sync byte
		memmove(input_buffer, input_ptr, remain);
		input_ptr = input_buffer;
		input_end = input_buffer + remain;
		buffer_file_offset += INPUT_BUFFER_SIZE - remain;
		// try to fill buffer 
		if (!input_eof) {
			result = fill_input_buffer();
			if (result != 0) {
				return result;
				}
			}
		remain = input_end - input_buffer;
		// handle as eof if not at least 3 packets to check
		if (remain < (TRANSPORT_PKT_SIZE * 3)) {
			return -1;
			}
		}
	}

if (buffer_sync_ok()) {
	if ((lost_sync_offset != 0) || (verbose_output)) {
		fprintf(stderr, "Achieved sync after %d bytes.\n",
				(int)((buffer_file_offset + (input_ptr - input_buffer)) - lost_sync_offset));
		}
	*packet = input_ptr;
	return 0;
	}
else {
	fprintf(stderr, "Failed to achieve sync after %uK bytes.\n",
			(sync_count + 1023) / 1024);
	return 1;
	}
}

int fill_input_buffer(void)
{
ssize_t read_length, read_actual;

// fill input buffer
// assumes input_eof has been checked
// and input_end is set to end of data to keep

// fill the buffer from the file
read_length = INPUT_BUFFER_SIZE;
read_length -= input_end - input_buffer;
read_actual = read(in_fd, input_end, read_length);
if (errno) {
	fprintf(stderr, "Error reading input file %s: %s\n", input_file, strerror(errno));
	return errno;
	}	
if (read_actual < read_length) {
	input_eof = 1;
	}
input_end += read_actual;

return 0;
}

int buffer_sync_ok(void)
{
	int remain;
	
	remain = input_end - input_ptr;
	
	if (remain >= (TRANSPORT_PKT_SIZE * 3)) {
		if ((*(input_ptr + (TRANSPORT_PKT_SIZE * 2))) != SYNC_BYTE){
			return 0;
			} 
		}

	if (remain >= (TRANSPORT_PKT_SIZE * 2)) {
		if ((*(input_ptr + TRANSPORT_PKT_SIZE)) != SYNC_BYTE){
			return 0;
			} 
		}

	if (remain >= TRANSPORT_PKT_SIZE) {
		return ((*input_ptr) == SYNC_BYTE);
		}

	return 0;
}

void print_help(char *myname)
{
	fprintf(stderr, "dtvstream version %s\n", VERSION);
	fprintf(stderr, "Format: %s [options]\n", myname); 
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "    -i input-file   Defaults to /dev/dtv\n");
	fprintf(stderr, "    -o output-file  Defaults to input-name.m2t\n");
#ifdef CHANNEL_TUNING
	fprintf(stderr, "    -c channel      tune to indicated TV channel (2 - 83)\n");
#endif
	fprintf(stderr, "    -p prog         select the given DTV program number\n");
	fprintf(stderr, "                    \"default\" for the default program\n");
	fprintf(stderr, "                    \"all\" to process all packets\n");
	fprintf(stderr, "    -m mins         stop after this many minutes (added to secs)\n");
	fprintf(stderr, "    -s secs         stop after this many seconds (added to mins)\n");
	fprintf(stderr, "    -a count        number of initial packets to analyze (default %d)\n", DEFAULT_SCAN_COUNT); 
	fprintf(stderr, "    -r    don't relace existing output file (default)\n");
	fprintf(stderr, "    -R    replace existing output file\n");
	fprintf(stderr, "    -q    just display stream program information\n");
	fprintf(stderr, "    -v    enable verbose output\n");
	fprintf(stderr, "    -d    enable debugging output (implies -v)\n");
	fprintf(stderr, "    -h    print this help text (also --help)\n");
}

int convert_to_int(char *str, int *convert_error)
{
	int result;
	char *strptr;
	
	result = (int)strtol(str, &strptr, 10);
	if (convert_error == 0) return result;
	
	if (*strptr == 0) {
		*convert_error = 0;
		}
	else {
		*convert_error = 1;
		}

	return result;
}

char * safe_strcat(char *dest, char *src, int src_copylen, int maxlen)
{
// len = declared length of dest
// strcat which appends only as much as fits
int src_len, dest_len, copylen;

if (src == 0) return dest;	// no src string
src_len = strlen(src);
if (src_len == 0) return dest;		// nothing to copy
if (src_copylen == 0) return dest;	// zero length requested 
if ((src_copylen > 0) && (src_copylen < src_len)) {
	src_len = src_copylen;			// use requested length if valid
	}

dest_len = strlen(dest);
copylen = maxlen - 1 - dest_len;
if (copylen <= 0) return dest; // already no room

if (copylen > src_len) copylen = src_len;  // no longer than source

memcpy(dest + dest_len, src, copylen);
dest[dest_len + copylen] = 0;

return dest;
}

void close_files(void)
{
if (in_open) {
	close(in_fd);
	in_open = 0;
	}

if (out_open) {
	close(out_fd);
	out_open = 0;
	}
}

